package dev.fitko.fitconnect.client.factory;

import com.nimbusds.jose.jwk.JWK;
import com.nimbusds.jose.jwk.RSAKey;
import dev.fitko.fitconnect.api.config.ApplicationConfig;
import dev.fitko.fitconnect.api.config.SchemaConfig;
import dev.fitko.fitconnect.api.config.SubscriberConfig;
import dev.fitko.fitconnect.api.exceptions.InitializationException;
import dev.fitko.fitconnect.api.exceptions.InvalidKeyException;
import dev.fitko.fitconnect.api.services.Sender;
import dev.fitko.fitconnect.api.services.Subscriber;
import dev.fitko.fitconnect.api.services.auth.OAuthService;
import dev.fitko.fitconnect.api.services.crypto.CryptoService;
import dev.fitko.fitconnect.api.services.crypto.MessageDigestService;
import dev.fitko.fitconnect.api.services.destination.DestinationService;
import dev.fitko.fitconnect.api.services.destination.KeyService;
import dev.fitko.fitconnect.api.services.events.EventLogService;
import dev.fitko.fitconnect.api.services.events.EventLogVerificationService;
import dev.fitko.fitconnect.api.services.events.SecurityEventService;
import dev.fitko.fitconnect.api.services.portal.SelfServicePortalService;
import dev.fitko.fitconnect.api.services.schema.SchemaProvider;
import dev.fitko.fitconnect.api.services.submission.SubmissionService;
import dev.fitko.fitconnect.api.services.validation.ValidationService;
import dev.fitko.fitconnect.client.SenderClient;
import dev.fitko.fitconnect.client.SubscriberClient;
import dev.fitko.fitconnect.core.SubmissionSender;
import dev.fitko.fitconnect.core.SubmissionSubscriber;
import dev.fitko.fitconnect.core.auth.DefaultOAuthService;
import dev.fitko.fitconnect.core.crypto.HashService;
import dev.fitko.fitconnect.core.crypto.JWECryptoService;
import dev.fitko.fitconnect.core.destination.DestinationApiService;
import dev.fitko.fitconnect.core.destination.PublicKeyService;
import dev.fitko.fitconnect.core.events.EventLogApiService;
import dev.fitko.fitconnect.core.events.EventLogVerifier;
import dev.fitko.fitconnect.core.events.SecurityEventTokenService;
import dev.fitko.fitconnect.core.http.ProxyConfig;
import dev.fitko.fitconnect.core.portal.SelfServicePortalApiService;
import dev.fitko.fitconnect.core.schema.SchemaResourceProvider;
import dev.fitko.fitconnect.core.submission.SubmissionApiService;
import dev.fitko.fitconnect.core.validation.DefaultValidationService;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.web.client.RestTemplate;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.text.ParseException;
import java.util.List;

/**
 * Factory that constructs clients for {@link Sender} and {@link Subscriber}.
 */
public final class ClientFactory {

    private static final Logger LOGGER = LoggerFactory.getLogger(ClientFactory.class);

    private static final String CONFIG_ENV_KEY_NAME = "FIT_CONNECT_CONFIG";
    private static final String SET_SCHEMA_DIR = "/set-schema";
    private static final String METADATA_SCHEMA_DIR = "/metadata-schema";

    private ClientFactory() {
    }

    /**
     * Create a new {@link SenderClient} that is automatically configured via the config.yml file.
     *
     * @return the sender client
     */
    public static SenderClient senderClient() {
        final ApplicationConfig config = loadConfig();
        return senderClient(config);
    }

    /**
     * Create a new {@link SenderClient} that is automatically configured via a provided {@link ApplicationConfig}.
     *
     * @return the sender client
     */
    public static SenderClient senderClient(final ApplicationConfig config) {
        LOGGER.info("Initializing sender client ...");
        return new SenderClient(getSender(config));
    }


    /**
     * Create a new {@link SubscriberClient} that is automatically configured via config.yml file.
     *
     * @return the subscriber client
     */
    public static SubscriberClient subscriberClient() {
        final ApplicationConfig config = loadConfig();
        return subscriberClient(config);
    }

    /**
     * Create a new {@link SubscriberClient} that is automatically configured via a provided {@link ApplicationConfig}.
     *
     * @return the subscriber client
     */
    public static SubscriberClient subscriberClient(final ApplicationConfig config) {
        LOGGER.info("Initializing subscriber client ...");
        final Subscriber subscriber = getSubscriber(config);
        final SubscriberConfig subscriberConfig = config.getSubscriberConfig();

        LOGGER.info("Reading private decryption key from {} ", subscriberConfig.getPrivateDecryptionKeyPath());
        final String privateKeyPath = readPath(subscriberConfig.getPrivateDecryptionKeyPath(), "Decryption Key");
        final RSAKey privateKey = readRSAKeyFromString(privateKeyPath);

        return new SubscriberClient(subscriber, privateKey);
    }

    private static Subscriber getSubscriber(final ApplicationConfig config) {
        final CryptoService cryptoService = getCryptoService();
        final RestTemplate restTemplate = getRestTemplate(config);
        final SchemaProvider schemaProvider = getSchemaProvider();
        final ValidationService validator = getValidationService(config, schemaProvider);
        final OAuthService authService = getSubscriberConfiguredAuthService(config, restTemplate);
        final EventLogService eventLogService = getEventLogService(config, restTemplate, authService);
        final SubmissionService submissionService = getSubmissionService(config, restTemplate, authService);
        final SecurityEventService setService = getSecurityEventTokenService(config, validator);
        final DestinationService destinationService = getDestinationService(config, restTemplate, authService);
        final SelfServicePortalService portalService = getSelfServicePortalService(config, restTemplate);
        final KeyService keyService = getKeyService(config, validator, destinationService, portalService);
        final EventLogVerificationService eventLogVerifier = getEventLogVerifier(keyService, validator);
        return new SubmissionSubscriber(submissionService, eventLogService, eventLogVerifier, cryptoService, validator, setService);
    }

    private static Sender getSender(final ApplicationConfig config) {
        final CryptoService cryptoService = getCryptoService();
        final RestTemplate restTemplate = getRestTemplate(config);
        final SchemaProvider schemaProvider = getSchemaProvider();
        final ValidationService validator = getValidationService(config, schemaProvider);
        final OAuthService authService = getSenderConfiguredAuthService(config, restTemplate);
        final DestinationService destinationService = getDestinationService(config, restTemplate, authService);
        final SelfServicePortalService portalService = getSelfServicePortalService(config, restTemplate);
        final KeyService keyService = getKeyService(config, validator, destinationService, portalService);
        final SubmissionService submissionService = getSubmissionService(config, restTemplate, authService);
        final EventLogService eventLogService = getEventLogService(config, restTemplate, authService);
        final EventLogVerificationService eventLogVerifier = getEventLogVerifier(keyService, validator);
        return new SubmissionSender(destinationService, submissionService, eventLogService, eventLogVerifier, cryptoService, validator, keyService);
    }

    private static OAuthService getSenderConfiguredAuthService(final ApplicationConfig config, final RestTemplate restTemplate) {
        final String clientId = config.getSenderConfig().getClientId();
        final String clientSecret = config.getSenderConfig().getClientSecret();
        return new DefaultOAuthService(restTemplate, clientId, clientSecret, config.getOAuthTokenEndpoint());
    }

    private static OAuthService getSubscriberConfiguredAuthService(final ApplicationConfig config, final RestTemplate restTemplate) {
        final String clientId = config.getSubscriberConfig().getClientId();
        final String clientSecret = config.getSubscriberConfig().getClientSecret();
        return new DefaultOAuthService(restTemplate, clientId, clientSecret, config.getOAuthTokenEndpoint());
    }

    private static SubmissionService getSubmissionService(final ApplicationConfig config, final RestTemplate restTemplate, final OAuthService authService) {
        return new SubmissionApiService(authService, restTemplate, config);
    }

    private static EventLogService getEventLogService(final ApplicationConfig config, final RestTemplate restTemplate, final OAuthService authService) {
        return new EventLogApiService(authService, restTemplate, config);
    }

    private static DestinationService getDestinationService(final ApplicationConfig config, final RestTemplate restTemplate, final OAuthService authService) {
        return new DestinationApiService(authService, restTemplate, config);
    }

    private static ValidationService getValidationService(final ApplicationConfig config, final SchemaProvider schemaProvider) {
        final MessageDigestService messageDigestService = getMessageDigestService();
        return new DefaultValidationService(config, messageDigestService, schemaProvider);
    }

    private static CryptoService getCryptoService() {
        final MessageDigestService messageDigestService = getMessageDigestService();
        return new JWECryptoService(messageDigestService);
    }

    private static RestTemplate getRestTemplate(final ApplicationConfig config) {
        final ProxyConfig proxyConfig = new ProxyConfig(config);
        return proxyConfig.proxyRestTemplate();
    }

    private static MessageDigestService getMessageDigestService() {
        return new HashService();
    }

    private static SecurityEventTokenService getSecurityEventTokenService(final ApplicationConfig config, final ValidationService validationService) {
        LOGGER.info("Reading private signing key from {} ", config.getSubscriberConfig().getPrivateSigningKeyPath());
        final String signingKey = readPath(config.getSubscriberConfig().getPrivateSigningKeyPath(), "Signing Key");
        final RSAKey rsaKey = readRSAKeyFromString(signingKey);
        return new SecurityEventTokenService(config, validationService, rsaKey);
    }

    private static SelfServicePortalService getSelfServicePortalService(final ApplicationConfig config, final RestTemplate restTemplate) {
        return new SelfServicePortalApiService(config, restTemplate);
    }

    private static KeyService getKeyService(final ApplicationConfig config, final ValidationService validator, final DestinationService destinationService, final SelfServicePortalService portalService) {
        return new PublicKeyService(config, destinationService, portalService, validator);
    }

    private static EventLogVerificationService getEventLogVerifier(final KeyService keyService, final ValidationService validationService) {
        return new EventLogVerifier(keyService, validationService);
    }

    private static SchemaProvider getSchemaProvider() {
        final List<String> metadataSchemaFiles = SchemaConfig.getMetadataSchemaFileNames(METADATA_SCHEMA_DIR);
        final List<String> setSchemaFiles = SchemaConfig.getSetSchemaFilePaths(SET_SCHEMA_DIR);
        return new SchemaResourceProvider(setSchemaFiles, metadataSchemaFiles);
    }

    private static RSAKey readRSAKeyFromString(final String key) {
        try {
            return RSAKey.parse(JWK.parse(key).toJSONObject());
        } catch (final ParseException | NullPointerException e) {
            LOGGER.error("No decryption key set, please check environment (config.yml) and provide a reference to the private decryption key");
            throw new InvalidKeyException("Key could not be parsed", e);
        }
    }

    private static String readPath(final String pathName, final String errorKey) {
        try {
            return Files.readString(Path.of(pathName));
        } catch (final IOException e) {
            LOGGER.error("Could not read {} from path '{}'. Please provide a valid path in your config.yml", errorKey, pathName);
            throw new InitializationException(e.getMessage(), e);
        }
    }

    private static Path readConfigFromEnvironment() {
        try {
            return Path.of(System.getenv(CONFIG_ENV_KEY_NAME));
        } catch (final NullPointerException e) {
            LOGGER.error("Environment variable {} could not be loaded", CONFIG_ENV_KEY_NAME);
            throw new InitializationException(e.getMessage(), e);
        }
    }

    private static ApplicationConfig loadConfig() {
        try {
            final Path configPath = readConfigFromEnvironment();
            final ApplicationConfig applicationConfig = ApplicationConfigLoader.loadConfig(configPath);
            LOGGER.info("Using sdk environment config `{}` ", applicationConfig.getActiveEnvironment().getName());
            return applicationConfig;
        } catch (final InitializationException e) {
            LOGGER.error("Config could not be loaded, please provide a 'config.yml' by setting the environment variable 'FIT_CONNECT_CONFIG'");
            throw new InitializationException(e.getMessage(), e);
        }
    }
}
